// $Id: threadqueue.cpp,v 1.1.1.1 2005/11/04 09:07:01 khlut Exp $
// Copyright 2005 Elphel, Inc.
//
// This file is part of GenReS.
//
//    GenReS is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    GenReS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with GenReS; if not, write to the Free Software
//    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

#include "nsCOMPtr.h"
#include "nsIServiceManagerUtils.h"
#include "nsIEventQueueService.h"
#include "nsIEventQueue.h"
#include <stdio.h>

static NS_DEFINE_CID( kEventQueueService, NS_EVENTQUEUESERVICE_CID );

extern "C" void PR_CALLBACK DestroyEvent(PLEvent* event) ;                                                                                   
extern "C" void * PR_CALLBACK HandleEvent(PLEvent* event) ;


extern "C" bool post_event_to_thread( PRThread *receiver, PLEvent *(*allocator)(void * owner, const void * msg), 
void * owner, const void * msg )
{
    nsresult rv; bool r=true;
    PLEvent * event;

    nsCOMPtr<nsIEventQueueService> eventQService = do_GetService(kEventQueueService, &rv);
    if(NS_FAILED(rv))	return false;
    nsCOMPtr<nsIEventQueue> eventQ;
    if(NS_FAILED( eventQService->GetThreadEventQueue(receiver,getter_AddRefs(eventQ)) ))	return false;
    eventQ->EnterMonitor();
    if((event=(*allocator)(owner, msg)))
      if(NS_FAILED( eventQ->PostEvent(event) )) {
	PL_DestroyEvent(event);
	r=false;
      }
    eventQ->ExitMonitor();

    return true;
}

extern "C" void ProcessOwnerEvents(PLEvent* event, void* data, PLEventQueue* queue) {
    if(event->owner == data) {
	PL_DequeueEvent(event, queue);
	PL_HandleEvent(event);
    }
}

extern "C" bool process_pending_events(void *owner)
{
    nsresult rv;

    // Get the event queue of the current thread...
    nsCOMPtr<nsIEventQueueService> eventQService = do_GetService(kEventQueueService, &rv);
    if(NS_FAILED(rv)) {
        printf("Unable to get event queue service!\n");
        return false;
    }

    nsCOMPtr<nsIEventQueue> eventQ;
     
    if(NS_FAILED( eventQService->GetThreadEventQueue(PR_GetCurrentThread(),getter_AddRefs(eventQ)) )) {
        printf("Unable to get thread queue!\n");
        return false;
    }
    
    PLEventQueue *PLEQ;
    eventQ->GetPLEventQueue(&PLEQ);
    PL_MapEvents(PLEQ,&ProcessOwnerEvents, owner);
    
    return true;
}
